//
// File:        BabelConfiguration.java
// Package:     gov.llnl.babel
// Release:     $Name: release-0-8-4 $
// Revision:    @(#) $Revision: 1.19 $
// Date:        $Date: 2003/03/18 17:46:27 $
// Description: Store configuration info shared by parsers and backend
// 
// Copyright (c) 2000-2002, The Regents of the University of Calfornia.
// Produced at the Lawrence Livermore National Laboratory.
// Written by the Components Team <components@llnl.gov>
// UCRL-CODE-2002-054
// All rights reserved.
// 
// This file is part of Babel. For more information, see
// http://www.llnl.gov/CASC/components/. Please read the COPYRIGHT file
// for Our Notice and the LICENSE file for the GNU Lesser General Public
// License.
// 
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU Lesser General Public License (as published by
// the Free Software Foundation) version 2.1 dated February 1999.
// 
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms and
// conditions of the GNU Lesser General Public License for more details.
// 
// You should have recieved a copy of the GNU Lesser General Public License
// along with this program; if not, write to the Free Software Foundation,
// Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

package gov.llnl.babel;

import gov.llnl.babel.symbols.RegexMatch;
import gov.llnl.babel.symbols.SymbolID;
import java.util.List;
import java.util.LinkedList;

/**
 * This class provides access to configuration information (including user
 * options) required by the parsers and backend.
 */
public class BabelConfiguration 
{
  /*
   * The basic configuration literals.
   */
  private static final String BASE_CLASS     = "SIDL.BaseClass";
  private static final String BASE_CLASSINFO = "SIDL.ClassInfo";
  private static final String BASE_CLASSINFOI= "SIDL.ClassInfoI";
  private static final String BASE_EXCEPTION = "SIDL.BaseException";
  private static final String BASE_INTERFACE = "SIDL.BaseInterface";
  private static final String MAKEFILE_NAME  = "babel.make";
  private static final int    MAXIMUM_ARRAY  = 4;

  /*
   * User options -- extractable from the command line via UserOptions.
   */
  private boolean d_generate_client        = false;
  private boolean d_generate_remote        = false;
  private boolean d_generate_server        = false;
  private boolean d_generate_stdlib        = false;
  private boolean d_generate_text          = false;
  private boolean d_generate_xml           = false;  // Deprecated
  private boolean d_package_subdirs        = false;
  private boolean d_parser_check           = false;
  private boolean d_suppress_timestamps    = false;
  private boolean d_comment_local_only     = false;
  private String  d_target_language        = "";
  private String  d_output_directory       = ".";
  private String  d_vpath_directory        = null;
  private StringBuffer d_repository_path   = new StringBuffer ("");
  private List    d_excluded_symbols       = new LinkedList();

  /*
   * My handle.
   */
  private static BabelConfiguration s_my_instance = null;


  /**
   * Since this is a singleton class, its constructor is protected.
   */
  protected BabelConfiguration() {
  }

  
  /**
   * Return the singleton instance of this class.
   */
  public static BabelConfiguration getInstance() {
    if (s_my_instance == null) {
      s_my_instance = new BabelConfiguration();
    }
    return s_my_instance;
  }

  
  /**
   * Return the full name of the root of the class type hierarchy.
   * This is the one class that does not have a parent class.
   */
  public static String getBaseClass() {
    return BASE_CLASS;
  }

  /**
   * Return the full name of the ClassInfo interface.
   */
  public static String getClassInfo() {
    return BASE_CLASSINFO;
  }

  /**
   * Return the full name of the class implementing ClassInfo interface.
   */
  public static String getClassInfoI() {
    return BASE_CLASSINFOI;
  }

  /**
   * The full name of the root of the interface type hierarchy.
   */
  public static String getBaseException() {
    return BASE_EXCEPTION;
  }

  /**
   * The full name of the base of the exception type hierarchy.
   * All exceptions must extend this type.
   */
  public static String getBaseInterface() {
    return BASE_INTERFACE;
  }

  /**
   * The full name of the file associated with generated makefile fragments.
   */
  public static String getMakefileName() {
    return MAKEFILE_NAME;
  }

  /**
   * Return the maximum array dimension.
   */
  public static int getMaximumArray() {
    return MAXIMUM_ARRAY;
  }

  /**
   * Return whether the specified symbol belongs to the SIDL namespace.
   */
  public static boolean isSIDLBaseClass(SymbolID id) { 
    boolean is_sidl = false;

    if (id != null) {
      String n = id.getFullName();
      if ((n != null) && (n.equals("SIDL") || n.startsWith("SIDL."))) {
         is_sidl = true;
      }
    }
    
    return is_sidl;
  }

  /**
   * Set the value of the server generation user option.
   */
  public void setGenerateServer(boolean generate) {
    d_generate_server = generate;
  }

  /**
   * Return the value of the server generation user option.
   */
  public boolean generateServer() {
    return d_generate_server;
  }
  
  /**
   * Set the value of the remote generation user option.
   */
  public void setGenerateRemote(boolean generate) {
    d_generate_remote = generate;
  }

  /**
   * Return the value of the remote generation user option.
   */
  public boolean generateRemote() {
    return d_generate_remote;
  }

  /**
   * Set the value of the client generation user option.
   */
  public void setGenerateClient(boolean generate) {
    d_generate_client = generate;
  }

  /**
   * Return the value of the client generation user option.
   */
  public boolean generateClient() {
    return d_generate_client;
  }
  
  /**
   * Set the value of the text generation user option.
   */
  public void setGenerateText(boolean generate) {
    d_generate_text = generate;
  }

  /**
   * Return the value of the text generation user option.
   */
  public boolean generateText() {
    return d_generate_text;
  }
  
  /**
   * Set the value of the XML generation user option. (Deprecated)
   */
  public void setGenerateXML(boolean generate) {
    d_generate_xml = generate;
  }

  /**
   * Return the value of the XML generation user option.  (Deprecated)
   */
  public boolean generateXML() {
    return d_generate_xml;
  }
  
  /**
   * Set the value of the SIDL stdlib generation user option.
   */
  public void setGenerateStdlib(boolean generate) {
    d_generate_stdlib = generate;
  }

  /**
   * Return the value of the SIDL stdlib generation user option.
   */
  public boolean generateStdlib() {
    return d_generate_stdlib;
  }
  
  /**
   * Set the value of the parser check user option.
   */
  public void setParseCheckOnly(boolean parse) {
    d_parser_check = parse;
  }

  /**
   * Return the value of the parser check user option.
   */
  public boolean parseCheckOnly() {
    return d_parser_check;
  }

  /**
   * Set the value of the comment local methods only option.
   */
  public void setCommentLocalOnly(boolean localOnly)
  {
    d_comment_local_only = localOnly;
  }

  /**
   * Return the value of the comment local methods only option.
   */
  public boolean getCommentLocalOnly()
  {
    return d_comment_local_only;
  }
  
  /**
   * Set the value of the timestamp suppression user option.
   */
  public void setSuppressTimestamps(boolean suppress) {
    d_suppress_timestamps = suppress;
  }

  /**
   * Return the value of the timestamp suppression user option.
   */
  public boolean suppressTimestamps() {
    return d_suppress_timestamps;
  }

  /**
   * Set the value of the make package subdirs user option.
   */
  public void setMakePackageSubdirs(boolean make_subdirs) {
    d_package_subdirs = make_subdirs;
  }
  
  /**
   * Return the value of the make package subdirs user option.
   */
  public boolean makePackageSubdirs() {
    return d_package_subdirs;
  }
  
  /**
   * Set the value of the output directory user option.
   */
  public void setOutputDirectory(String dir) {
    d_output_directory = dir;
  }

  /**
   * Return the value of the output directory user option.
   */
  public String getOutputDirectory() {
    return d_output_directory;
  }
  
  /**
   * Set the value of the vpath directory user option.
   */
  public void setVPathDirectory(String dir) {
    d_vpath_directory = dir;
  }

  /**
   * Return the value of the vpath directory user option.
   */
  public String getVPathDirectory() {
    return d_vpath_directory;
  }
  
  /**
   * Set the value of the target language user option.
   */
  public void setTargetLanguage(String lang) {
    d_target_language = lang;
  }

  /**
   * Return the value of the target language user option.
   */
  public String getTargetLanguage() {
    return d_target_language;
  }
  
  /**
   * Adds a new entry to the repository path.
   */
  public void addToRepositoryPath(String new_path) {
    if ( d_repository_path != null ) {
      if ( d_repository_path.length() == 0 ) {
        d_repository_path.insert (0, new_path);
      } else {
        d_repository_path.insert (0, new_path + ";");
      }
    } else {
      d_repository_path = new StringBuffer(new_path);
    }
  }

  /**
   * Return the value of the repository path.
   */
  public String getRepositoryPath() {
    return d_repository_path.toString();
  }

  /**
   * Add another regular expression to the list of excluded
   * regular expressions.
   */
  public void addExcluded(RegexMatch rm)
  {
    d_excluded_symbols.add(rm);
  }

  /**
   * Return the list of {@link gov.llnl.babel.symbols.RegexMatch} objects
   * that should be excluded from code generation.
   */
  public List getExcludedList()
  {
    return d_excluded_symbols;
  }
  
}
