//
// File:	CommandLineDriver.java
// Package:	gov.llnl.babel
// Release:	$Name: release-0-8-4 $
// Revision:	$Revision: 1.34 $
// Modified:	$Date: 2003/01/14 18:41:59 $
// Description:	main driver for the babel IDL compiler
//
// Copyright (c) 2000-2002, The Regents of the University of Calfornia.
// Produced at the Lawrence Livermore National Laboratory.
// Written by the Components Team <components@llnl.gov>
// UCRL-CODE-2002-054
// All rights reserved.
// 
// This file is part of Babel. For more information, see
// http://www.llnl.gov/CASC/components/. Please read the COPYRIGHT file
// for Our Notice and the LICENSE file for the GNU Lesser General Public
// License.
// 
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU Lesser General Public License (as published by
// the Free Software Foundation) version 2.1 dated February 1999.
// 
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms and
// conditions of the GNU Lesser General Public License for more details.
// 
// You should have recieved a copy of the GNU Lesser General Public License
// along with this program; if not, write to the Free Software Foundation,
// Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

package gov.llnl.babel;

import gov.llnl.babel.backend.BuildGenerator;
import gov.llnl.babel.backend.CodeGenerationException;
import gov.llnl.babel.backend.CodeGenerationFactory;
import gov.llnl.babel.backend.CodeGenerator;
import gov.llnl.babel.backend.FileListener;
import gov.llnl.babel.backend.FileManager;
import gov.llnl.babel.parsers.sidl.Parser;
import gov.llnl.babel.parsers.sidl.SIDLException;
import gov.llnl.babel.repository.Repository;
import gov.llnl.babel.repository.RepositoryException;
import gov.llnl.babel.repository.RepositoryFactory;
import gov.llnl.babel.symbols.Symbol;
import gov.llnl.babel.symbols.SymbolID;
import gov.llnl.babel.symbols.SymbolNotFoundException;
import gov.llnl.babel.symbols.SymbolRedefinitionException;
import gov.llnl.babel.symbols.SymbolTable;
import gov.llnl.babel.symbols.RegexMatch;
import gov.llnl.babel.url.URLUtilities;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.StringTokenizer;


/**
 * This class is the main driver for running Babel from the command line.
 */
public class CommandLineDriver
{
  /**
   * The command line parser.
   */
  private static UserOptions d_option_parser = UserOptions.getInstance();

  /**
   * The user options and basic babel configuration information.
   */
  private static BabelConfiguration s_babel_config = BabelConfiguration.
                                                       getInstance();

  /**
   * The file generation manager.
   */
  private FileManager d_file_manager = null;

  /**
   * The symbol table
   */
  private SymbolTable d_symbol_table = null;

  /**
   * The handle to the singleton instance of this class.
   */
  private static CommandLineDriver s_my_instance = null;


  /**
   * The protected singleton constructor instantiates the options parser,
   * <code>UserOptions</code>.
   *
   * @see UserOptions
   */
  protected CommandLineDriver() {
  } 

  /**
   * Return the singleton instance of this class.
   */
  private static CommandLineDriver getInstance() {
    if ( s_my_instance == null ) {
      s_my_instance = new CommandLineDriver();
    }
    return s_my_instance;
  } 


  /**
   * Extract the options from the command line arguments and perform
   * associated set up as appropriate.  
   *
   * @see BabelConfiguration
   */
  private void processCommandline( String args[] )
  { 
    /*
     * First extract the options from the arguments and determine if
     * we should proceed with the compilation.
     */
    int firstUnusedArg = d_option_parser.parseCommandlineOptions(s_babel_config,
                                                                 args);

    if ( !d_option_parser.canProceed() ) {
      System.exit(1);
    }

    /*
     * Now continue processing the remainder of the arguments and
     * create associated processing elements as needed.
     */

    resolveRepositoryPaths();

    setupFileManager();

    try { 
      resolveRemainingArgs( args, firstUnusedArg );
    } catch ( CollectionException ex ) { 
      System.err.println("Babel: Error: when trying to resolve remaining args...");
      System.err.println( ex.getMessage() );
      System.exit(1);
    }

    /*
     * Stop now if only checking parsing.
     */
    if ( s_babel_config.parseCheckOnly() ) {
      System.exit(0);
    }
  } 


  /**
   * Create <code>Repository</code> instances based on the user-specified
   * repository path and add them to the <code>SymbolTable</code> singleton.
   * 
   * @see SymbolTable
   */
  private void resolveRepositoryPaths() 
  { 
    String repoPath               = s_babel_config.getRepositoryPath();
    RepositoryFactory the_factory = RepositoryFactory.getInstance();

    d_symbol_table                = SymbolTable.getInstance();

    for( StringTokenizer tokenizer = new StringTokenizer( repoPath, ";" );
         tokenizer.hasMoreTokens(); ) 
    {
      try 
      { 
        String path = tokenizer.nextToken();
        if ( (path == null) || (path.equals("")) || (path.equals("null")) ) { 
          continue;
        } 
        d_symbol_table.addSymbolResolver(the_factory.createRepository(path));
      } catch ( RepositoryException ex ) { 
        System.err.println("Babel: Warning: When creating repository...");
        System.err.println( ex.getMessage() );
	ex.printStackTrace( System.err );
      }
    }
  }


  /**
   * Initialize the file generation manager with configuration-specific
   * information as needed.  
   *
   * @see BabelConfiguration
   * @see FileManager
   */
  private void setupFileManager()
  { 
    d_file_manager   = FileManager.getInstance();
    if ( s_babel_config == null ) {
      System.err.println("Babel: Warning: configuration was not populated prior to "
                         + "setting up the file manager.");
      s_babel_config = BabelConfiguration.getInstance();
    }

    try { 
      d_file_manager.setFileGenerationRootDirectory(s_babel_config.
                                                      getOutputDirectory());
    } catch ( CodeGenerationException ex ) { 
      System.err.println("Babel: Warning: when trying to set output directory..." );
      System.err.println( ex.getMessage() );
      ex.printStackTrace( System.err );
    }
    
    d_file_manager.setJavaStylePackageGeneration(s_babel_config.
                                                   makePackageSubdirs());
    try { 
      d_file_manager.setVPathDirectory( s_babel_config.getVPathDirectory() );
    } catch ( CodeGenerationException ex ) { 
      System.err.println("Babel: Warning: when trying to set VPATH directory..." );
      System.err.println( ex.getMessage() );
      ex.printStackTrace( System.err );
    }
  }

  private CollectionException addToCollection(CollectionException ce,
                                              String uri,
                                              Exception ex)
  {
    if (ce == null) {
      ce = new CollectionException(uri, ex);
    }
    else {
      ce.addException(uri, ex);
    }
    return ce;
  }


  /**
   * Process the remainder of the command line, assuming the contents
   * reflect one or more SIDL files.  
   */
  private void resolveRemainingArgs( String args[], 
                                     int    firstUnusedArg ) 
    throws CollectionException
  { 
    /*
     * First parse the remaining arguments.
     */
    CollectionException ce = null;
    for( int i=firstUnusedArg; i<args.length; i++ ) 
    {
      try 
      { 
        File f = new File( args[i] );
        if ( f.exists() && f.isFile() && f.canRead() ) { 
          /*
           * Parse the URL for the readable file.
           */
          parseURL( f.toURL().toString() );
        } else if ( f.exists() && f.isDirectory() ) {
          /*
           * If it's a directory, assume user meant a symbol, so resolve it.
           */
          resolveSymbolByString( args[i] );
        } else if ( args[i].indexOf(':') != -1 ) { 
          /*
           * Try it as a URL if there's a colon.
           */
          parseURL( URLUtilities.expandURL(args[i]) );
        } else if ( resolveSymbolByString( args[i] ) == null ) { 
          /*
           * Otherwise, it must be tried as a symbol.
           */
          throw new CodeGenerationException( "Error : \"" + args[i] 
                                  + "\" fails to resolve as a symbol or file.");
        }
      }
      catch ( Exception ex ) { 
        ce = addToCollection(ce, args[i], ex);
      }
    } // end for loop

    /*
     * Now resolve all the symbols by first handling those by SIDL base class.
     */
    if ( d_symbol_table == null ) {
      System.err.println("Babel: Warning: symbol table was not populated prior to "
                         + "symbol resolution.");
      d_symbol_table = SymbolTable.getInstance();
    }

    try {
      d_symbol_table.resolveSymbol("SIDL");
      d_symbol_table.resolveAllReferences();
    }
    catch ( Exception ex ) {
      ce = addToCollection(ce, null, ex);
    } 
    if (ce != null) {
      throw ce;
    }
  }


  /**
   * Open and parse contents of the specified URL.
   */
  private static void parseURL( String url ) 
    throws SIDLException, MalformedURLException, IOException
  { 
    URL         u      = new URL( url );
    InputStream s = null;
    try {
      s = u.openStream();
      Parser      parser = new Parser(s);

      System.out.println("Babel: Parsing URL \"" + url + "\"...");

      parser.setSourceURL( url );
      String[] warnings = parser.beginParse();

      if ( warnings != null ) { 
        for ( int w=0; w<warnings.length; w++ ) { 
          System.err.println("Babel: Warning: " + warnings[w] );
        }
      }
    }
    finally {
      if (s != null) s.close();
    }
  }


  /**
   * Try to parse string as a symbol.  Return null if it fails.
   */
  private Symbol resolveSymbolByString( String symbol_string ) 
    throws NumberFormatException, SymbolRedefinitionException 
  {
    
    int    split        = symbol_string.indexOf("-v");
    String symbol_name  = ( split>0 && split<symbol_string.length() ) ?
                            symbol_string.substring(0,split) : symbol_string;
    String version_name = ( split>0 && (split+2)<symbol_string.length() ) ?
                            symbol_string.substring(split+2) : null;
    Symbol symbol       = null;

    if ( (version_name == null) || (version_name == "") ) { 
      symbol = d_symbol_table.resolveSymbol( symbol_name );
    } else { 
      SymbolID id = new SymbolID( symbol_name, 
                            new gov.llnl.babel.symbols.Version( version_name ));
      symbol = d_symbol_table.resolveSymbol( id );
    }

    if ( symbol != null ) { 
      System.out.println("Babel: Resolved symbol \"" + symbol_string + "\"...");
    }
    return symbol;
  }


  /**
   * Generate the appropriate source code based on user configuration 
   * parameters.
   */
  private void generateSourceCode()
  {
    if ( d_symbol_table == null ) {
      System.err.println("Babel: Warning: symbol table was not populated prior to "
                         + "generating source code");
      d_symbol_table = SymbolTable.getInstance();
    }
    Set symbols = d_symbol_table.getSymbolNames();

    if ( s_babel_config == null ) {
      System.err.println("Babel: Warning: configuration was not populated prior to "
                         + "generating the source code");
      s_babel_config = BabelConfiguration.getInstance();
    }
    symbols = excludeSymbols(symbols);

    if ( s_babel_config.generateXML() ) {	// Deprecated!
      System.err.println("Babel: Warning: The --xml option is now deprecated, "
                         + "please start using the --text=xml option instead");
      generateXMLCode(symbols);
    } else {
      String targetLanguage = s_babel_config.getTargetLanguage();
      if ( s_babel_config.generateText() ) { 
        if ( targetLanguage.equals("xml") ) {
          generateXMLCode(symbols);
        } else {
          try {
            d_symbol_table.resolveAllParents();
            generateCode(symbols, "text", targetLanguage);
          }
          catch (SymbolNotFoundException snfe) {
            System.err.println("Babel: Error: SIDL backend cannot find all parent packages.");
            System.err.println(snfe.getMessage());
          }
        }
      } else {
        BuildGenerator mg = getBuildGen(targetLanguage);
        if (mg instanceof FileListener) {
          d_file_manager.addListener((FileListener)mg);
        }
        if ( s_babel_config.generateClient() ) {
          generateClientSource(symbols);
          generateMakefiles(mg);
        } else if ( s_babel_config.generateServer() ) {
          generateServerSource(symbols);
          generateMakefiles(mg);
        } else if ( s_babel_config.generateRemote() ) {
          generateRemoteSource(symbols);
          generateMakefiles(mg);
        }
      }
    }
  }


  /**
   * Generate the XML repository code.
   */
  private void generateXMLCode( Set symbols )
  {
    Repository repository;
    if ( s_babel_config == null ) {
      System.err.println("Babel: Warning: configuration was not populated prior to "
                         + "generating the XML code");
      s_babel_config = BabelConfiguration.getInstance();
    }

    try { 
      repository = RepositoryFactory.getInstance().createRepository(
                                       s_babel_config.getOutputDirectory());
      if ( repository != null ) { 
        repository.writeSymbols(symbols);
      }
    } 
    catch ( RepositoryException ex ) 
    { 
      System.err.println("Babel: Error: when creating output repository...");
      System.err.println( ex.getMessage() );
      ex.printStackTrace( System.err );
      repository = null;
    }
  }

  private void generateCode(Set symbols, String type, String language)
  {
    CodeGenerator gen = getCodeGen(language, type);
    if (gen != null) {
      if (gen.getUserSymbolsOnly()) {
        symbols = filterSymbols(symbols);
      }
      try {
        gen.generateCode(symbols);
      }
      catch (CodeGenerationException ex) {
        System.err.println
          ("Babel: Warning: when trying to generate " +
           type + " code...");
        System.err.println( ex.getMessage() );
	ex.printStackTrace( System.err );
      }
    }
    else {
      System.err.println
        ("Babel: Warning: when generating code...unable to generate the " +
         type + " sources in " + language + ".");
    }
  }


  /**
   * Generate the server source code.
   */
  private void generateServerSource( Set symbols )
  {
    /*
     * First generate the client code.
     */
     generateClientSource(symbols);

    /*
     * Now generate the server-specific code.
     */
    String targetLanguage = s_babel_config.getTargetLanguage();
    generateCode(symbols, "skel", targetLanguage);
  }


  /**
   * Generate the client source code.
   */
  private void generateClientSource( Set symbols )
  {
    String targetLanguage = s_babel_config.getTargetLanguage();
    generateCode(symbols, "ior", "ior");
    generateCode(symbols, "stub", targetLanguage);
  }


  /**
   * Generate the remote client source code.
   */
  private void generateRemoteSource( Set symbols )
  {
    System.err.println("Babel: Warning: when generating code..."
                       + "process to generate the remote sources undefined.");
  }

  private BuildGenerator getBuildGen(String myLang)
  {
    BuildGenerator bg = 
      CodeGenerationFactory.getInstance().getBuildGenerator(myLang);
    if (bg == null){
      System.err.println("Babel: Warning: getBuildGen( " + myLang +
                         ") returned null.");
    }
    return bg;
  }

  /**
   * Instantiate the appropriate code generator associated with the desired
   * language and mode.
   *
   * @see CodeGenerator
   */
  private CodeGenerator getCodeGen( String myLang, String mode )
  {
    CodeGenerator cg;

    cg = CodeGenerationFactory.getInstance().getCodeGenerator( myLang, mode );
    if ( cg == null ) { 
      System.err.println("Babel: Warning: getCodeGen( " + myLang + ", " + mode +
                         ") returned null.");
    }
    return cg;
  }

  private static boolean excludeSymbol(SymbolID id)
  {
    Iterator i = s_babel_config.getExcludedList().iterator();
    while (i.hasNext()) {
      if (((RegexMatch)i.next()).match(id)) return false;
    }
    return true;
  }

  private static Set excludeSymbols(Collection symbols)
  {
    Set result = new HashSet(symbols.size());
    Iterator i = symbols.iterator();
    while (i.hasNext()) {
      SymbolID id = (SymbolID)i.next();
      if (excludeSymbol(id)) result.add(id);
    }
    return result;
  }

  private static boolean filterSymbol(SymbolID id)
  {
    return !BabelConfiguration.isSIDLBaseClass(id);
  }


  /**
   * Depending on the configuration settings, return either all symbols or
   * just the user symbols, those symbols outside the SIDL namespace.
   */
  private static Set filterSymbols(Set symbols) 
  {
    Set result;
    if (s_babel_config.generateStdlib()) {
      result = symbols;
    }
    else {
      result = new HashSet(symbols.size());
      Iterator i = symbols.iterator();
      while (i.hasNext()) {
        SymbolID id = (SymbolID)i.next();
        if (filterSymbol(id)) result.add(id);
      }
    }
    return result;
  }


  /**
   * Generate the appropriate makefile fragments.
   */
  private void generateMakefiles(BuildGenerator mg)
  {
    /*
     * Now generate the fragments.
     */
    try { 
      mg.createAll();
    } catch ( IOException ex ) { 
      System.out.println("Warning: when trying to generate makefiles...");
      System.out.println(ex.getMessage());
    }
  }


  /****************************************************************
   * Main babel entry point.
   */
  public static void main( String args[] )
  {
    try {
      /*
       * Set up the configuration parameters based on command line arguments.
       */
      CommandLineDriver myDriver = getInstance();
      myDriver.processCommandline(args);
    
      /*
       * Generate the source code.
       */
      myDriver.generateSourceCode();
    }
    catch (Throwable t) {
      System.err.println("Babel: Unexplained exception...");
      System.err.println(t.getMessage());
      t.printStackTrace();
      Runtime.getRuntime().exit(1);
    }
    finally {
      /* workaround for JNI bug in Linux 1.3.1 JVM */
      Runtime.getRuntime().exit(0);
    }
  }
} 
